import logging
from uuid import UUID

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db.models import Q
from drf_spectacular.utils import OpenApiExample, OpenApiParameter, extend_schema, extend_schema_view
from elasticsearch.exceptions import NotFoundError
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import NotFound, PermissionDenied
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import (
    Corpus,
    Element,
    Entity,
    EntityLink,
    EntityRole,
    Transcription,
    TranscriptionEntity,
)
from arkindex.documents.serializers.elements import ElementEntitiesSerializer, ElementSlimSerializer
from arkindex.documents.serializers.entities import (
    BaseEntitySerializer,
    CreateEntityRoleErrorResponseSerializer,
    EntityCreateSerializer,
    EntityLinkCreateSerializer,
    EntityLinkSerializer,
    EntityRoleSerializer,
    EntitySerializer,
    TranscriptionEntityDetailsSerializer,
    TranscriptionEntitySerializer,
)
from arkindex.project.elastic import ESEntity
from arkindex.project.mixins import ACLMixin, CorpusACLMixin
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.project.triggers import reindex_start
from arkindex.users.models import Role

logger = logging.getLogger(__name__)


@extend_schema(tags=['entities'])
@extend_schema_view(
    get=extend_schema(operation_id='ListCorpusRoles', description='List all roles of a corpus'),
    post=extend_schema(
        description='Create a new entity role',
        responses={
            200: EntityRoleSerializer,
            400: CreateEntityRoleErrorResponseSerializer
        },
        examples=[OpenApiExample(
            status_codes=['400'],
            response_only=True,
            name="role-exists",
            value={'id': "55cd009d-cd4b-4ec2-a475-b060f98f9138", 'corpus': ["Role already exists in this corpus"]},
            description="Role already exists."
        )]
    )
)
class CorpusRoles(CorpusACLMixin, ListCreateAPIView):
    """
    List all roles in a corpus
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = EntityRoleSerializer
    queryset = EntityRole.objects.none()

    def get_queryset(self):
        return EntityRole.objects \
            .filter(corpus=self.get_corpus(self.kwargs['pk'])) \
            .order_by('parent_name', 'child_name')

    def perform_create(self, serializer):
        data = self.request.data
        if EntityRole.objects.filter(
            parent_name=data['parent_name'],
            child_name=data['child_name'],
            parent_type=data['parent_type'],
            child_type=data['child_type'],
            corpus_id=self.request.parser_context['kwargs']['pk']
        ).exists():
            raise serializers.ValidationError({
                'corpus': ['Role already exists in this corpus'],
                'id': self.request.parser_context['kwargs']['pk']
            })
        super().perform_create(serializer)


@extend_schema(tags=['entities'])
@extend_schema_view(
    patch=extend_schema(description='Partially update an entity'),
    put=extend_schema(description='Update an entity'),
    delete=extend_schema(description='Delete an entity')
)
class EntityDetails(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get all information about an entity
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = EntitySerializer

    def get_queryset(self):
        return Entity.objects \
            .select_related('corpus') \
            .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
            .prefetch_related(
                'parents__role',
                'children__role',
                'parents__child',
                'parents__parent',
                'children__parent',
                'children__child',
                'corpus'
            )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if self.request.method not in permissions.SAFE_METHODS and not self.has_access(obj.corpus, Role.Contributor.value):
            raise PermissionDenied(detail='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        if settings.ARKINDEX_FEATURES['search']:
            # Try to delete indexed entity if possible
            try:
                es_entity = ESEntity.get(id=instance.id.hex)
                es_entity.delete()
            except NotFoundError:
                pass
            except Exception as e:
                logger.error(f"Failed to delete ES index entity {instance.id}: {e}")

        instance.delete()


@extend_schema_view(get=extend_schema(operation_id='ListEntityElements', tags=['entities']))
class EntityElements(ListAPIView):
    """
    Get all elements that have a link with the entity
    """
    serializer_class = ElementSlimSerializer
    # For OpenAPI type discovery: an entity's ID is in the path
    queryset = Entity.objects.none()

    def get_queryset(self):
        pk = self.kwargs['pk']
        metadata_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                metadatas__entity_id=pk
            ) \
            .select_related('type', 'corpus') \
            .prefetch_related('metadatas__entity', 'zone__image__server')
        transcription_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                transcriptions__transcription_entities__entity_id=pk
            ).select_related('type', 'corpus') \
            .prefetch_related('metadatas__entity', 'zone__image__server')
        return metadata_elements.union(transcription_elements) \
            .order_by('name', 'type')


@extend_schema_view(post=extend_schema(
    operation_id='CreateEntity',
    responses={201: EntitySerializer},
    tags=['entities'],
))
class EntityCreate(CreateAPIView):
    """
    Create a new entity
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityCreateSerializer

    def perform_create(self, serializer):
        name = serializer.validated_data['name']
        type = serializer.validated_data['type']
        corpus = serializer.validated_data['corpus']
        metas = serializer.validated_data['metas'] if 'metas' in serializer.data else None
        worker_version = serializer.validated_data['worker_version']
        return Entity.objects.create(
            name=name,
            type=type,
            corpus=corpus,
            metas=metas,
            worker_version=worker_version
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        entity = EntitySerializer(self.perform_create(serializer))
        # Index entity into Elasticsearch
        reindex_start(entity=entity.instance, entities=True)
        headers = self.get_success_headers(serializer.data)
        return Response(entity.data, status=status.HTTP_201_CREATED, headers=headers)


@extend_schema_view(post=extend_schema(operation_id='CreateEntityLink', tags=['entities']))
class EntityLinkCreate(CreateAPIView):
    """
    Create a new link between two entities with a role
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityLinkCreateSerializer


@extend_schema_view(post=extend_schema(
    operation_id='CreateTranscriptionEntity',
    tags=['entities'],
    parameters=[
        OpenApiParameter(
            'id',
            type=UUID,
            location=OpenApiParameter.PATH,
            description='ID of the transcription to link an entity to.',
            required=True,
        )
    ]
))
class TranscriptionEntityCreate(CreateAPIView):
    """
    Link an existing Entity to a given transcription with its position
    """
    permission_classes = (IsVerified, )
    serializer_class = TranscriptionEntitySerializer
    # For OpenAPI type discovery: a transcription's ID is in the path
    queryset = Transcription.objects.none()

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if not self.request:
            # OpenAPI cannot pass request context
            return context

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        tr = Transcription.objects.using('default').filter(
            id=self.kwargs.get('pk'),
            element__corpus__in=Corpus.objects.writable(self.request.user)
        ).first()
        return {**context, 'transcription': tr}

    def perform_create(self, serializer):
        data = serializer.validated_data
        if TranscriptionEntity.objects.filter(**data).exists():
            raise serializers.ValidationError({
                '__all__': ['Entity is already linked to this transcription at this position']
            })
        try:
            TranscriptionEntity(**data).clean()
        except ValidationError as e:
            raise serializers.ValidationError(serializers.as_serializer_error(e))
        super().perform_create(serializer)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListTranscriptionEntities',
        tags=['entities'],
        parameters=[
            OpenApiParameter(
                'worker_version',
                type=UUID,
                description='Only include entities created by a specific worker version',
                required=False,
            )
        ]
    )
)
class TranscriptionEntities(ListAPIView):
    """
    List existing entities linked to a specific transcription
    """
    serializer_class = TranscriptionEntityDetailsSerializer
    # For OpenAPI type discovery: a transcription's ID is in the path
    queryset = Transcription.objects.none()

    def get_queryset(self):
        filters = {
            'transcription__element__corpus__in': Corpus.objects.readable(self.request.user),
            'transcription_id': self.kwargs['pk'],
        }
        if 'worker_version' in self.request.query_params:
            try:
                worker_version_id = UUID(self.request.query_params['worker_version'])
            except (TypeError, ValueError):
                raise serializers.ValidationError({'worker_version': ['Invalid UUID']})

            if not WorkerVersion.objects.filter(id=worker_version_id).exists():
                raise serializers.ValidationError({'worker_version': ['This worker version does not exist.']})

            filters['entity__worker_version_id'] = worker_version_id

        return TranscriptionEntity.objects \
            .filter(**filters) \
            .order_by('offset') \
            .prefetch_related('entity')


@extend_schema_view(
    get=extend_schema(
        operation_id='ListElementEntities',
        tags=['entities'],
        parameters=[
            OpenApiParameter(
                'worker_version',
                type=UUID,
                description='Only include entities created by a specific worker version',
                required=False,
            )
        ]
    )
)
class ElementEntities(RetrieveAPIView):
    """
    List all entities linked to an element's transcriptions and metadata
    """
    serializer_class = ElementEntitiesSerializer

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        return Element.objects.filter(corpus__in=corpora).select_related('type').only('id', 'type__slug', 'name')


@extend_schema_view(
    get=extend_schema(
        operation_id='ListCorpusEntities',
        tags=['entities'],
    )
)
class CorpusEntities(CorpusACLMixin, ListAPIView):
    """
    List all entities in a corpus.
    """
    serializer_class = BaseEntitySerializer

    def get_queryset(self):
        return self.get_corpus(self.kwargs['pk']).entities.order_by('name', 'id')


@extend_schema_view(get=extend_schema(operation_id='ListElementLinks', tags=['entities']))
class ElementLinks(CorpusACLMixin, ListAPIView):
    """
    List all links where parent and child are linked to the element.\n\n
    Requires a **guest** access to the element corpus
    """
    serializer_class = EntityLinkSerializer

    def get_queryset(self):
        try:
            element = Element.objects.select_related('corpus').get(id=self.kwargs['pk'])
        except Element.DoesNotExist:
            raise NotFound

        if not self.has_read_access(element.corpus):
            raise PermissionDenied(detail='You do not have access to this element.')

        # Load entities linked by transcriptions
        entities_tr = Entity.objects.filter(transcriptions__element_id=element.id).prefetch_related('transcriptions')

        # Load entities linked by metadatas
        entities_meta = Entity.objects.filter(metadatas__element_id=element.id).prefetch_related('metadatas')

        # Now load all links belonging to those entities
        # It's several times faster to combine the queries in the final one
        # than combining them at the lower level (entities is slower than entities_tr + entities_meta)
        # We need to support cross references between transcriptions & metadata entities
        return EntityLink.objects.filter(
            Q(parent__in=entities_tr, child__in=entities_tr)
            | Q(parent__in=entities_tr, child__in=entities_meta)
            | Q(parent__in=entities_meta, child__in=entities_tr)
            | Q(parent__in=entities_meta, child__in=entities_meta)
        ).select_related('role', 'child', 'parent').order_by('parent__name')
