import datetime
import itertools
import json
import logging

from django.conf import settings
from django.db.models import prefetch_related_objects
from SolrClient import SolrClient
from SolrClient.exceptions import SolrError
from teklia_toolbox.time import Timer

from arkindex.documents.models import Element

logger = logging.getLogger(__name__)

solr = SolrClient(settings.SOLR_API_URL)


class Indexer(object):

    # Chunk of element processed (can generate many more documents)
    elements_chunk_size = 100

    # Chunk of documents sent to Solr
    solr_chunk_size = 200
    solr_num_shards = 1
    solr_options = {'commit': True}
    solr_type_fields = [
        {'name': 'uuid', 'class': 'solr.UUIDField'},
        # Update string analyzer to support case-insensitive searching
        {
            'name': 'string',
            'class': 'solr.TextField',
            'analyzer': {
                'tokenizer': {'class': 'solr.KeywordTokenizerFactory'},
                'filters': [{'class': 'solr.LowerCaseFilterFactory'}]
            }
        }
    ]
    solr_fields = [
        # Parent fields
        {'name': 'parent_id', 'indexed': False, 'required': True, 'type': 'uuid'},
        {'name': 'parent_name', 'indexed': False, 'required': True, 'type': 'string'},
        {'name': 'parent_type', 'indexed': False, 'required': True, 'type': 'string'},
        # Element fields
        {'name': 'element_id', 'indexed': False, 'required': True, 'type': 'uuid'},
        {'name': 'element_text', 'indexed': True, 'required': True, 'type': 'string'},
        {'name': 'element_type', 'indexed': True, 'required': True, 'type': 'string'},
        {'name': 'element_worker', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'element_image', 'indexed': False, 'required': False, 'type': 'string'},
        # Transcription fields
        {'name': 'transcription_id', 'indexed': False, 'required': False, 'type': 'uuid'},
        {'name': 'transcription_text', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'transcription_confidence', 'indexed': True, 'required': False, 'type': 'pfloat'},
        {'name': 'transcription_worker', 'indexed': True, 'required': False, 'type': 'string'},
        # Classification fields
        {'name': 'classification_id', 'indexed': False, 'required': False, 'type': 'uuid'},
        {'name': 'classification_name', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'classification_confidence', 'indexed': True, 'required': False, 'type': 'pfloat'},
        {'name': 'classification_worker', 'indexed': True, 'required': False, 'type': 'string'},
        # Metadata fields
        {'name': 'metadata_id', 'indexed': False, 'required': False, 'type': 'uuid'},
        {'name': 'metadata_name', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'metadata_text', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'metadata_type', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'metadata_worker', 'indexed': True, 'required': False, 'type': 'string'},
        # Entity fields
        {'name': 'entity_id', 'indexed': False, 'required': False, 'type': 'uuid'},
        {'name': 'entity_text', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'entity_type', 'indexed': True, 'required': False, 'type': 'string'},
        {'name': 'entity_worker', 'indexed': True, 'required': False, 'type': 'string'}
    ]

    def __init__(self, corpus_id):
        self.corpus_id = corpus_id
        self.collection_name = f'project-{self.corpus_id}'

    def setup(self):
        """
        Create collection in Solr
        """
        if not solr.collections.exists(self.collection_name):
            logger.info(f'Creating collection {self.collection_name}')
            solr.collections.create(self.collection_name, self.solr_num_shards)
        else:
            logger.info(f'Collection {self.collection_name} already exists')

        logger.info(f'Updating field types for {self.collection_name}')
        # Disable SolrClient logging
        logging.disable(logging.ERROR)
        for type_field in self.solr_type_fields:
            # SolrClient does not support field type management
            params = {
                'method': 'POST',
                'endpoint': solr.schema.schema_endpoint,
                'collection': self.collection_name,
            }
            try:
                solr.transport.send_request(**params, data=json.dumps({'add-field-type': type_field}))
            except SolrError as e:
                if f"Field type '{type_field['name']}' already exists." not in str(e):
                    raise e
                solr.transport.send_request(**params, data=json.dumps({'replace-field-type': type_field}))
        # Restore logging
        logging.disable(logging.NOTSET)

        logger.info(f'Updating fields for {self.collection_name}')
        # Disable SolrClient logging
        logging.disable(logging.INFO)
        for field in self.solr_fields:
            if not solr.schema.does_field_exist(self.collection_name, field['name']):
                solr.schema.create_field(self.collection_name, field)
            else:
                solr.schema.replace_field(self.collection_name, field)
        # Restore logging
        logging.disable(logging.NOTSET)

    def drop_index(self):
        """
        Drop an existing collection
        """
        solr.delete_doc_by_query(self.collection_name, '*:*', **self.solr_options)
        logger.info(f'Dropped index for {self.collection_name}')

    def get_elements(self):
        """
        Use a SQL query to keep the link between parent and child across element paths (a parent is also linked to itself)
        This query lists the parent and child information (id, name and type name)
        The WHERE clause limits the selection to a single corpus and limits the parent elements to indexable elements
        """

        query = f"""
        SELECT parent.id as parent_id,
            parent.name as parent_name,
            parenttype.display_name as parent_type,
            element.id as id,
            element.name as name,
            elementtype.display_name as type_name
        FROM documents_element as element
        LEFT OUTER JOIN documents_elementpath as elementpath ON (element.id = elementpath.element_id)
        LEFT OUTER JOIN documents_element as parent ON (
            elementpath.path @> ARRAY[parent.id]
            OR element.id = parent.id
        )
        INNER JOIN documents_elementtype parenttype ON (parent.type_id = parenttype.id)
        INNER JOIN documents_elementtype elementtype ON (element.type_id = elementtype.id)
        WHERE element.corpus_id = '{self.corpus_id}' AND parenttype.indexable
        """
        return Element.objects.raw(query)

    def elements_chunk(self, elements):
        element_iterator = elements.iterator()
        while True:
            chunk = tuple(itertools.islice(element_iterator, self.elements_chunk_size))
            if not chunk:
                return
            yield chunk

    def hash_worker(self, worker_version):
        if not worker_version:
            return

        worker = worker_version.worker
        return f'{worker.id.hex[:10]}-{worker.name}'

    def build_documents(self, elements):
        BUILD_METHODS = [self.build_transcriptions, self.build_classifications, self.build_metadatas, self.build_entities]
        documents = []
        for element in elements:
            document = self.build_element(element)
            documents.append(document)
            for method in BUILD_METHODS:
                documents += method(element, document)
        return documents

    def build_element(self, element):
        return {
            'parent_id': str(element.parent_id),
            'parent_name': element.parent_name,
            'parent_type': element.parent_type,
            'element_id': str(element.id),
            'element_text': element.name,
            'element_type': element.type_name,
            'element_worker': self.hash_worker(element.worker_version),
            'element_image': element.zone.thumbnail_url if element.zone else None
        }

    def build_transcriptions(self, element, document):
        return [
            dict(document, **{
                'transcription_id': str(transcription.id),
                'transcription_text': transcription.text,
                'transcription_confidence': transcription.confidence,
                'transcription_worker': self.hash_worker(transcription.worker_version)
            }) for transcription in element.transcriptions.all()
        ]

    def build_classifications(self, element, document):
        return [
            dict(document, **{
                'classification_id': str(classification.id),
                'classification_name': classification.ml_class.name,
                'classification_confidence': classification.confidence,
                'classification_worker': self.hash_worker(classification.worker_version)
            }) for classification in element.classifications.all()
        ]

    def build_metadatas(self, element, document):
        return [
            dict(document, **{
                'metadata_id': str(metadata.id),
                'metadata_name': metadata.name,
                'metadata_text': metadata.value,
                'metadata_type': metadata.type.value,
                'metadata_worker': self.hash_worker(metadata.worker_version)
            }) for metadata in element.metadatas.all()
        ]

    def build_entities(self, element, document):
        entities = [entity for transcription in element.transcriptions.all() for entity in transcription.entities.all()]
        entities += [metadata.entity for metadata in element.metadatas.all() if metadata.entity]
        return [
            dict(document, **{
                'entity_id': str(entity.id),
                'entity_text': entity.name,
                'entity_type': entity.type.value,
                'entity_worker': self.hash_worker(entity.worker_version)
            }) for entity in entities
        ]

    def index(self):
        """
        Insert items into Solr
        Process elements with an indexable type and their children by chunck
        For each chunk:
            - load their dependencies (transcriptions, classifications...)
            - serialize items in a Solr document
            - send the documents to Solr
        """
        elements = self.get_elements()

        total_elements, total_documents = 0, 0
        retrieve_time, build_time, index_time = datetime.timedelta(0), datetime.timedelta(0), datetime.timedelta(0)

        for elements_chunk in self.elements_chunk(elements):
            nb_elements = len(elements_chunk)
            total_elements += nb_elements
            logger.debug(f'Processing {nb_elements} elements...')
            # Retrieve elements from db
            with Timer() as t:
                # Element
                prefetch_related_objects(elements_chunk, 'zone__image__server')
                prefetch_related_objects(elements_chunk, 'worker_version__worker')
                # Transcriptions
                prefetch_related_objects(elements_chunk, 'transcriptions')
                prefetch_related_objects(elements_chunk, 'transcriptions__worker_version__worker')
                # Classifications
                prefetch_related_objects(elements_chunk, 'classifications')
                prefetch_related_objects(elements_chunk, 'classifications__worker_version__worker')
                # Metadatas
                prefetch_related_objects(elements_chunk, 'metadatas')
                prefetch_related_objects(elements_chunk, 'metadatas__worker_version__worker')
                # Entities
                prefetch_related_objects(elements_chunk, 'transcriptions__entities')
                prefetch_related_objects(elements_chunk, 'transcriptions__entities__worker_version__worker')
                prefetch_related_objects(elements_chunk, 'metadatas__entity')
                prefetch_related_objects(elements_chunk, 'metadatas__entity__worker_version__worker')
            retrieve_time += t.delta
            logger.debug(f'Retrieved {nb_elements} elements')

            # Build Solr documents
            with Timer() as t:
                documents = self.build_documents(elements_chunk)
                nb_documents = len(documents)
                total_documents += nb_documents
            build_time += t.delta
            logger.debug(f'Built {nb_documents} Solr documents')

            # Index documents into Solr
            with Timer() as t:
                for i in range(0, nb_documents, self.solr_chunk_size):
                    solr.index(self.collection_name, documents[i:i + self.solr_chunk_size], **self.solr_options)
            index_time += t.delta
            logger.debug(f'Indexed {nb_documents} documents into Solr')

            result = solr.query(self.collection_name, {'q': '*:*'})
            logger.info(f'Currently {result.get_num_found()} documents in Solr')

        logger.info(f'Retrieved {total_elements} elements in {retrieve_time}')
        logger.info(f'Built {total_documents} Solr documents in {build_time}')
        logger.info(f'Indexed {total_documents} documents into Solr in {index_time}')

        result = solr.query(self.collection_name, {'q': '*:*'})
        logger.info(f'Currently {result.get_num_found()} documents in Solr')
