from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, Element
from arkindex.images.models import POLYGON_MAX_POINTS, ImageServer
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestCreateElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.page_type = cls.corpus.types.get(slug='page')
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.element = Element.objects.get(name='Volume 1, page 2r')
        cls.image = ImageServer.objects.local.images.create(
            path="kingdom/far/away",
            status=S3FileStatus.Checked,
            width=42,
            height=42,
        )
        cls.worker_version = WorkerVersion.objects.get(worker__slug='dla')

    def make_create_request(self, name='default', corpus=None, elt_type='volume', **options):
        request = {
            'path': reverse('api:elements-create'),
            'data': {
                'type': elt_type,
                'name': name,
                'corpus': corpus or str(self.corpus.id),
            },
            'format': 'json',
        }
        # Optional parameters
        request['data'].update(options)
        return request

    def test_create_element_not_verified(self):
        self.assertEqual(
            self.client.post(**self.make_create_request('no')).status_code,
            status.HTTP_403_FORBIDDEN
        )

    def test_create_element_volume(self):
        # Create a Volume
        self.client.force_login(self.user)
        request = self.make_create_request('my new volume')
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        data = response.json()
        self.assertIn('id', data)
        volume = Element.objects.get(id=data['id'])
        # Assert the fully serialized response matches the created volume
        self.assertDictEqual(
            data,
            {
                'id': str(volume.id),
                'name': volume.name,
                'type': volume.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': volume.thumbnail.s3_url,
                'worker_version': None,
                'classifications': [],
                'metadata': None,
                'corpus': {
                    'id': str(volume.corpus.id),
                    'name': volume.corpus.name,
                    'public': volume.corpus.public
                },
                'zone': None
            }
        )
        self.assertEqual(volume.name, 'my new volume')
        self.assertEqual(volume.type, self.volume_type)

    def test_create_volume_page(self):
        # Add a page with an Image to an existing volume
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams',
            image=str(self.image.id),
        )
        with self.assertNumQueries(17):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        # Assert the fully serialized response matches the created page (exclude image zone)
        del data['zone']
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'name': page.name,
                'type': page.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': None,
                'worker_version': None,
                'classifications': [],
                'metadata': None,
                'corpus': {
                    'id': str(page.corpus.id),
                    'name': page.corpus.name,
                    'public': page.corpus.public
                }
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams')
        self.assertEqual(page.type, self.page_type)

    def test_create_volume_act(self):
        # Create an volume child of type Act
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            name='Castle story',
            elt_type='act'
        )
        with self.assertNumQueries(14):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)

    def test_create_element_worker_version(self):
        # Create an element with a worker version
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='Castle story',
            elt_type='act',
            worker_version=str(self.worker_version.id),
        )
        with self.assertNumQueries(9):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.worker_version, self.worker_version)

    def test_create_element_polygon_max_size(self):
        # Element polygon is limited to POLYGON_MAX_POINTS points
        polygon = [(0, 0), *((i, i) for i in range(POLYGON_MAX_POINTS - 1, -1, -1))]
        self.assertEqual(len(polygon), POLYGON_MAX_POINTS + 1)
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams again',
            image=str(self.image.id),
            polygon=polygon
        )
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(), {'polygon': ['Ensure this field has no more than 300 elements.']}
        )

    def test_create_element_polygon(self):
        # Create an element with a polygon to an existing volume
        polygon = [[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]]
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams again',
            image=str(self.image.id),
            polygon=polygon
        )
        with self.assertNumQueries(17):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        # Assert the fully serialized response matches the created element (exclude zone image field)
        del data['zone']['image']
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'name': page.name,
                'type': page.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': None,
                'worker_version': None,
                'classifications': [],
                'metadata': None,
                'corpus': {
                    'id': str(page.corpus.id),
                    'name': page.corpus.name,
                    'public': page.corpus.public
                },
                'zone': {
                    'id': str(page.zone.id),
                    'polygon': polygon,
                    'url': str(page.zone.url)
                }
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams again')
        self.assertEqual(page.type, self.page_type)

    def test_create_parent_different_corpus(self):
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        new_corpus.types.create(slug='act')
        request = self.make_create_request(
            parent=str(self.vol.id),
            corpus=str(new_corpus.id),
            elt_type='act'
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'corpus': ['Parent and child must be in the same corpus']
        })

    def test_create_filter_type(self):
        """
        Ensure types are filtered by corpus
        """
        self.assertTrue(self.corpus.types.filter(slug='volume').exists())
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        new_vol_type = new_corpus.types.create(slug='volume', display_name='Volume', folder=True)
        request = self.make_create_request(
            corpus=str(new_corpus.id),
            elt_type='volume',
            name='something',
        )
        with self.assertNumQueries(7):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_vol = new_corpus.elements.get()
        self.assertEqual(new_vol.type, new_vol_type)
        self.assertEqual(new_vol.type.corpus, new_corpus)

    def test_create_wrong_type(self):
        self.assertFalse(self.corpus.types.filter(slug='kartoffelsalad').exists())
        self.client.force_login(self.user)
        request = self.make_create_request(
            corpus=str(self.corpus.id),
            elt_type='kartoffelsalad',
            name='something',
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ["Type 'kartoffelsalad' not found in corpus Unit Tests"],
        })

    def test_create_image_not_checked(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.status = S3FileStatus.Error
        self.image.save()
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertEqual(element.zone.image.status, S3FileStatus.Error)

    def test_create_image_zero_dimensions(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.width = 0
        self.image.height = 0
        self.image.status = S3FileStatus.Unchecked
        self.image.save()
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'image': ['This image does not have valid dimensions.']})

    def test_create_element_slim_output_simple(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='slim output !',
        )
        request['path'] += '?slim_output=true'
        with self.assertNumQueries(7):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })

    def test_create_element_slim_output(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='slim output !',
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id),
        )
        request['path'] += '?slim_output=true'
        with self.assertNumQueries(16):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })
