import json
import random
from pathlib import Path

from django.db.utils import IntegrityError
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.images.managers import PolygonTooLargeError
from arkindex.images.models import POLYGON_MAX_POINTS, Image, Zone
from arkindex.project.gis import ensure_linear_ring
from arkindex.project.tests import FixtureAPITestCase

SAMPLES = Path(__file__).absolute().parent / 'samples'


class TestImagePolygon(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.image = Image.objects.get(path='img1')
        cls.large_polygon_json = json.loads((SAMPLES / 'large_polygon.json').read_text())
        cls.large_polygon = ensure_linear_ring(cls.large_polygon_json)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.element = cls.corpus.elements.exclude(zone=None).first()

    def setUp(self):
        super().setUp()
        self.client.force_login(self.superuser)

    def test_polygon_num_pts_constraint(self):
        """
        Polygon maximum allowed number of points is empirically set to POLYGON_MAX_POINTS
        """
        # Create a polygon with POLYGON_MAX_POINTS + 1 points
        polygon = [(0, 0), *((i, i) for i in range(POLYGON_MAX_POINTS - 1, -1, -1))]
        self.assertEqual(len(polygon), POLYGON_MAX_POINTS + 1)
        self.assertEqual(polygon[0], polygon[-1])
        with self.assertRaisesRegex(IntegrityError, '"images_zone" violates check constraint "zone_polygon_size"'):
            Zone(image=self.image, polygon=polygon).save()

    def test_polygon_POLYGON_MAX_POINTS_pts(self):
        """
        Ensure a Polygon of POLYGON_MAX_POINTS points with a rather good entropy can be saved
        """
        random.seed(1337)
        polygon = [(0, 0), *((random.randint(0, 100000), i) for i in range(1, POLYGON_MAX_POINTS - 1)), (0, 0)]
        self.assertEqual(len(polygon), POLYGON_MAX_POINTS)
        self.assertEqual(polygon[0], polygon[-1])
        Zone(image=self.image, polygon=polygon).save()

    def test_objects_create_error(self):
        with self.assertRaises(PolygonTooLargeError):
            Zone.objects.create(image=self.image, polygon=self.large_polygon)

    def test_objects_update_error(self):
        self.assertTrue(Zone.objects.filter(image=self.image).exists())
        with self.assertRaises(PolygonTooLargeError):
            Zone.objects.filter(image=self.image).update(polygon=self.large_polygon)

    def test_zones_create_error(self):
        with self.assertRaises(PolygonTooLargeError):
            self.image.zones.create(polygon=self.large_polygon)

    def test_zones_update_error(self):
        self.assertTrue(self.image.zones.exists())
        with self.assertRaises(PolygonTooLargeError):
            self.image.zones.update(polygon=self.large_polygon)

    def test_zone_save_error(self):
        zone = Zone(image=self.image, polygon=self.large_polygon)
        with self.assertRaises(PolygonTooLargeError):
            zone.save()

    def test_zone_save_update_error(self):
        zone = self.image.zones.first()
        zone.polygon = self.large_polygon
        with self.assertRaises(PolygonTooLargeError):
            zone.save()

    def test_create_element_polygon_error(self):
        response = self.client.post(
            reverse('api:elements-create'),
            data={
                'corpus': str(self.corpus.id),
                'type': 'text_line',
                'name': 'large polygon',
                'image': str(self.image.id),
                'polygon': self.large_polygon_json,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR)
        self.assertDictEqual(response.json(), {
            'detail': 'This polygon has exceeded the database size limits. Consider simplifying it.'
        })

    def test_update_element_polygon_error(self):
        response = self.client.put(
            reverse('api:element-retrieve', kwargs={'pk': self.element.id}),
            data={
                'type': 'text_line',
                'name': 'large polygon',
                'image': str(self.image.id),
                'polygon': self.large_polygon_json,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR)
        self.assertDictEqual(response.json(), {
            'detail': 'This polygon has exceeded the database size limits. Consider simplifying it.'
        })

    def test_partial_update_element_polygon_error(self):
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': self.element.id}),
            data={
                'polygon': self.large_polygon_json,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR)
        self.assertDictEqual(response.json(), {
            'detail': 'This polygon has exceeded the database size limits. Consider simplifying it.'
        })

    def test_create_elements_polygon_error(self):
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': self.element.id}),
            data={
                'worker_version': str(self.worker_version.id),
                'elements': [
                    {
                        'name': 'good element',
                        'type': 'page',
                        'polygon': [[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]],
                    },
                    {
                        'name': 'bad element',
                        'type': 'text_line',
                        'polygon': self.large_polygon_json,
                    }
                ]
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR)
        self.assertDictEqual(response.json(), {
            'detail': 'This polygon has exceeded the database size limits. Consider simplifying it.'
        })

    def test_create_element_transcriptions_polygon_error(self):
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.element.id}),
            data={
                'element_type': 'text_line',
                'worker_version': str(self.worker_version.id),
                'transcriptions': [
                    {
                        'polygon': [[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]],
                        'text': 'a',
                        'confidence': .1,
                    },
                    {
                        'polygon': self.large_polygon_json,
                        'text': 'b',
                        'confidence': .2,
                    }
                ]
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR)
        self.assertDictEqual(response.json(), {
            'detail': 'This polygon has exceeded the database size limits. Consider simplifying it.'
        })
