from unittest.mock import call, patch

import requests
import responses
from django.contrib.gis.geos import LinearRing

from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureTestCase


class TestImagePerformCheck(FixtureTestCase):

    def setUp(self):
        self.imgsrv.refresh_from_db()
        self.img = self.imgsrv.images.create(
            path='checkimg',
            width=1000,
            height=1000,
            status=S3FileStatus.Checked,
        )
        self.zone = self.img.zones.create(polygon=LinearRing((10, 10), (10, 100), (100, 100), (100, 10), (10, 10)))

    @responses.activate
    def test_status(self):
        """
        Test Image.perform_check fails on a HTTP error code
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', status=418)
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaises(requests.HTTPError):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_empty(self):
        """
        Test Image.perform_check fails when the server returns nothing
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', status=200)
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaisesRegex(ValueError, 'Server returned an invalid JSON document'):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_json(self):
        """
        Test Image.perform_check fails when the server returns invalid JSON
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', body='<blink>oh no</blink>')
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaisesRegex(ValueError, 'Server returned an invalid JSON document'):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_required_items(self):
        """
        Test Image.perform_check fails when required items are missing from the JSON payload
        """
        base_payload = {
            '@id': 'http://server/checkimg',
            'width': 42,
            'height': 42,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        }
        for key in base_payload:
            payload = base_payload.copy()
            del payload[key]
            responses.add(responses.GET, 'http://server/checkimg/info.json', json=payload)
            with self.assertRaisesRegex(AssertionError, 'Missing required properties'):
                self.img.perform_check(raise_exc=True, save=False)
            self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_check_id(self):
        """
        Test Image.perform_check ensures the image is on the expected server
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://wrongserver/checkimg',
            'width': 42,
            'height': 42,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaisesRegex(AssertionError, 'Image id does not start with server url'):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_update_fields(self):
        """
        Test Image.perform_check updates the path, width and height
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg111',
            'width': 9000,
            'height': 8000,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.assertEqual(self.img.path, 'checkimg')
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        self.assertEqual(self.img.server.max_width, None)
        self.img.status = S3FileStatus.Unchecked

        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.path, 'checkimg111')
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((80, 80), (80, 800), (800, 800), (800, 80), (80, 80)))

    @responses.activate
    def test_resize_ignore_no_size(self):
        """
        Test Image.perform_check updates the width and height
        and ignores resizing when the previous image size was 0×0 to avoid ZeroDivisionErrors
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg',
            'width': 9000,
            'height': 8000,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.img.width = 0
        self.img.height = 0
        self.assertEqual(self.img.server.max_width, None)
        self.img.status = S3FileStatus.Unchecked

        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((10, 10), (10, 100), (100, 100), (100, 10), (10, 10)))

    @responses.activate
    def test_max_size(self):
        """
        Test Image.perform_check handles the IIIF maxWidth and maxHeight optional properties
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg',
            'width': 9000,
            'height': 8000,
            'profile': [
                'http://iiif.io/api/image/2/level2.json',
                {
                    "maxWidth": 4500,
                    "maxHeight": 10000,
                }
            ],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.server.max_height = 10000
        self.img.server.max_width = 4500
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((80, 80), (80, 800), (800, 800), (800, 80), (80, 80)))

    @responses.activate
    def test_max_size_error(self):
        """
        Test Image.perform_check handles compares the received maxWidth and maxHeight against the ImageServer's
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg',
            'width': 9000,
            'height': 8000,
            'profile': [
                'http://iiif.io/api/image/2/level2.json',
                {
                    "maxWidth": 4500,
                    "maxHeight": 10000,
                }
            ],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.server.max_height = None
        self.img.server.max_width = None
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        with self.assertRaises(AssertionError) as context:
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(
            str(context.exception),
            'Image server maximum width/height (None×None) does not match the actual maximum width/height (4500×10000). '
            'Please ask an instance administrator to update the IIIF server configuration.'
        )

    @responses.activate
    def test_max_height_deduced(self):
        """
        Test Image.perform_check assumes maxHeight is equal to maxWidth
        when maxHeight is missing and maxWidth is defined
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg',
            'width': 1000,  # width < maxWidth
            'height': 8000,  # height > maxWidth
            'profile': [
                'http://iiif.io/api/image/2/level2.json',
                {
                    "maxWidth": 4000,
                }
            ],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.server.max_height = 4000
        self.img.server.max_width = 4000
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 8000)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((10, 10), (10, 100), (100, 100), (100, 10), (10, 10)))

    @responses.activate
    def test_harvard_non_conform(self):
        """
        Test Image.perform_check falls back to properties from the Image Information object
        when they are not defined in the profile description object or when it is missing.
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg',
            'width': 9001,
            'maxWidth': 9000,
            'height': 9001,
            'maxHeight': 8999,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })

        self.img.status = S3FileStatus.Unchecked
        self.img.server.max_height = 8999
        self.img.server.max_width = 9000
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.width, 9001)
        self.assertEqual(self.img.height, 9001)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((90, 90), (90, 900), (900, 900), (900, 90), (90, 90)))

    @responses.activate
    def test_iiif_v1(self):
        """
        Test Image.perform_check supports a IIIF Image API 1.x payload
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@context': 'http://library.stanford.edu/iiif/image-api/1.1/context.json',
            '@id': 'http://server/checkimg111',
            'width': 9000,
            'height': 8000,
            'profile': 'http://library.stanford.edu/iiif/image-api/1.1/compliance.html#level0'
        })
        self.assertEqual(self.img.path, 'checkimg')
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        self.img.status = S3FileStatus.Unchecked

        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.path, 'checkimg111')
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((80, 80), (80, 800), (800, 800), (800, 80), (80, 80)))

    @responses.activate
    def test_profile_uri_validation(self):
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@context': 'http://library.stanford.edu/iiif/image-api/1.1/context.json',
            '@id': 'http://server/checkimg111',
            'width': 9000,
            'height': 8000,
            'profile': 'lol',
        })

        with patch('arkindex.images.models.logger') as logger_mock:
            self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.path, 'checkimg111')
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((80, 80), (80, 800), (800, 800), (800, 80), (80, 80)))

        self.assertEqual(logger_mock.warning.call_count, 1)
        self.assertEqual(logger_mock.warning.call_args, call('Invalid IIIF profile URI'))

    @responses.activate
    def test_server_config_check(self):
        """
        Ensure Image.perform_check warns about incorrect maximum width/height settings on the ImageServer
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg',
            'width': 9000,
            'height': 8000,
            'profile': [
                'http://iiif.io/api/image/2/level2.json',
                {
                    "maxWidth": 4500,
                    "maxHeight": 10000,
                }
            ],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.server.max_height = 10
        self.img.server.max_width = 10
        with self.assertRaisesRegex(AssertionError, 'maximum width/height'):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((10, 10), (10, 100), (100, 100), (100, 10), (10, 10)))

    @responses.activate
    def test_server_config_check_none(self):
        """
        Ensure Image.perform_check ignores other profile settings on the ImageServer
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg',
            'width': self.img.width,
            'height': self.img.height,
            'profile': [
                'http://iiif.io/api/image/2/level2.json',
                {
                    "supports": [
                        "sizeByW",
                    ]
                }
            ],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Checked)

        self.zone.refresh_from_db()
        self.assertEqual(self.zone.polygon.coords, ((10, 10), (10, 100), (100, 100), (100, 10), (10, 10)))

    @responses.activate
    def test_perform_check_failing_dimensions(self):
        """
        Test Image.perform_check fails with non-positive dimensions
        """
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg111',
            'width': 0,
            'height': 0,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaisesRegex(AssertionError, 'Image dimensions must be strictly positive integers'):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_perform_check(self):
        responses.add(responses.GET, 'http://server/checkimg/info.json', json={
            '@id': 'http://server/checkimg111',
            'width': 100,
            'height': 100,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Checked)

    @responses.activate
    def test_perform_check_weird_url(self):
        responses.add(responses.GET, 'http://server/checkimg/info.json', status=500)
        responses.add(responses.GET, 'http://server/checkimg/?IIIF=/info.json', json={
            '@id': 'http://server/checkimg111',
            'width': 100,
            'height': 100,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.img.path = "checkimg/?IIIF="
        self.img.status = S3FileStatus.Unchecked
        self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Checked)

    @responses.activate
    def test_perform_check_urls_ko(self):
        responses.add(responses.GET, 'http://server/checkimg/info.json', status=500)
        responses.add(responses.GET, 'http://server/checkimg/?IIIF=/info.json', status=500)
        self.img.path = "checkimg/?IIIF="
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaises(requests.HTTPError):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)
