from collections.abc import Iterable
from itertools import groupby

from django.contrib.gis.db.models.fields import BaseSpatialField, LineStringField
from django.contrib.gis.db.models.functions import GeoFuncMixin, GeomOutputGeoFunc
from django.contrib.gis.db.models.functions import MemSize as MemSizeFunc
from django.contrib.gis.db.models.functions import NumPoints as NumPointsFunc
from django.contrib.gis.db.models.lookups import EqualsLookup
from django.contrib.gis.geos import GEOSGeometry, LineString, Polygon
from django.db.models import BooleanField, Transform


def ensure_linear_ring(value):
    if isinstance(value, str):
        value = GEOSGeometry(value, srid=0)

    if isinstance(value, LineString):
        coords = value.coords
    elif isinstance(value, Iterable) and not isinstance(value, GEOSGeometry):
        coords = list(value)
    else:
        raise TypeError('A LinearRing, LineString, iterable of coordinates, '
                        'or a WKT representation of a LINESTRING or LINEARRING is required.')

    if not coords:
        # The deduplication would crash without a single point
        raise ValueError('At least 4 distinct points are required.')

    # PostGIS has a ST_RemoveRepeatedPoints method to deduplicate points, which is implemented by PostGIS itself
    # instead of the usual pattern of just acting as a layer between GEOS or GDAL and Postgres, like what is done
    # for ST_Normalize; this means it is not provided in the Django GEOS bindings, so we reimplement it on the
    # Python side using itertools.groupby.
    coords = next(zip(*groupby(coords)))
    if len(coords) < 4:
        raise ValueError('At least 4 distinct points are required.')

    polygon = Polygon(coords)
    polygon.normalize()
    return polygon.exterior_ring


class SimplifyPreserveTopology(GeomOutputGeoFunc):
    """
    Simplify a LineString, MultiLineString, Polygon or MultiPolygon. Does nothing with other geometries.
    Unlike ST_Simplify, this will avoid creating invalid geometries that would otherwise break some constraints.

    Docs: https://postgis.net/docs/ST_SimplifyPreserveTopology.html
    """
    arity = 2

    def __init__(self, expression, tolerance, **kwargs):
        super().__init__(
            expression,
            self._handle_param(tolerance, 'tolerance', (int, float)),
            **kwargs,
        )


@LineStringField.register_lookup
class IsClosed(GeoFuncMixin, Transform):
    """
    Returns a boolean: True when a LineString's first and last point are the same.
    https://postgis.net/docs/ST_IsClosed.html
    """
    lookup_name = 'isclosed'
    output_field = BooleanField()


@BaseSpatialField.register_lookup
class NumPoints(NumPointsFunc, Transform):
    """
    Number of points in a geometry.
    Provided by Django, but modified for use as a Transform in a queryset:
    Zone.objects.filter(polygon__numpoints=42)

    https://postgis.net/docs/ST_NumPoints.html
    https://docs.djangoproject.com/en/3.1/ref/contrib/gis/functions/#numpoints
    """
    lookup_name = 'numpoints'


@BaseSpatialField.register_lookup
class MemSize(MemSizeFunc, Transform):
    """
    Size, in bytes, of a geometry.
    This lookup is not used anymore except for migrations.

    https://postgis.net/docs/ST_MemSize.html
    https://docs.djangoproject.com/en/3.1/ref/contrib/gis/functions/#memsize
    """
    lookup_name = 'memsize'


# Django uses the `~=` (same-as) operator to compare geometries by default;
# however, in PostGIS >=1.5, this causes a comparison by bounding boxes only.
# This line rewrites the `__exact` lookup (the default lookup) to be `ST_Equals`,
# which is already provided by Django and properly compares vertex by vertex.
BaseSpatialField.register_lookup(EqualsLookup, 'exact')
